/*
 * lesslog - PostgreSQL archive log size reduction utility
 *
 * Version 1.4.1, for PostgreSQL 8.4.x
 *
 *    Copyright(C) 2010 Nippon Telegraph and Telephone Corporation
 *
 *    Contact: koichi.szk@gmail.com
 */

/*
 * log.c
 *    Get names of log records
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "postgres.h"
#include "access/clog.h"
#include "access/gin.h"
#include "access/gist_private.h"
#include "access/htup.h"
#include "access/multixact.h"
#include "access/nbtree.h"
#include "access/transam.h"
#include "access/xact.h"
#include "access/xlog_internal.h"
#include "catalog/pg_control.h"
#include "commands/dbcommands.h"
#include "commands/tablespace.h"
#include "storage/bufpage.h"
#include "commands/sequence.h"
#include "utils/relmapper.h"
#include "storage/standby.h"

#include "access/xlog.h"
#include "access/xlog_internal.h"

#include "log.h"
#include "debug.h"

static FILE *logfile = NULL;
static uint32 segment_id;

#define LOG_ENV "LESSLOG_LOG"
#define APPEND_ENV "LESSLOG_APPEND"

#define INFO_MASK 0xF0

#define CK_LOGGING do{if (logfile==NULL) return;}while(0)

static void get_record_name(int xl_rmid, 
                            int info, char **rmid_name, char **rec_name);
void finish_log(void);
void init_xlogdump(const char *progname, const char *logfname);
void log_BlkBoundary();


void init_xlogdump(const char *progname, const char *logfname)
{

    if (strcmp(logfname, "stdout") == 0) {
        logfile = stdin;
    }
    else if (strcmp(logfname, "stderr") == 0) {
        logfile = stderr;
    }
    else if ((logfile = fopen(logfname, "w")) == NULL) {
        fprintf(stderr, "Pg_lesslog: %s: cannot open log file \"%s\", %s\n",
                progname, logfname, strerror(errno));
    }
    CK_LOGGING;
    fprintf(logfile, "\n\n**** %s: XlogDump **************************\n\n",
            progname);
    return;
}

void init_log(const char *progname)
{
    char *logfname;
    char *logappend;

    if (((logfname = getenv(LOG_ENV)) == NULL) || !logfname[0]) {
        return;
    }
    logappend = getenv(APPEND_ENV);
    if (logappend == NULL) {
        logappend = "w";
    }
    if ((logfile = fopen(logfname, strcmp(logappend, "a") == 0 ? "a" : "w")) == NULL) {
        fprintf(stderr, "Pg_lesslog: %s: cannot open log file \"%s\", %s\n",
                progname, logfname, strerror(errno));
    }
    else {
        fprintf(logfile, "\n\n**** %s: Log **************************\n\n",
            progname);
    }
    return;
}


void finish_log()
{
    if (logfile) {
        fclose(logfile);
    }
    logfile = NULL;
}

void log_segment_id(const char *filename)
{
    TimeLineID tli;
    uint32 xlogid;
    char *p;

    CK_LOGGING;

    if (filename == NULL) {
        segment_id = 0;
        return;
    }
    p = strrchr(filename, '/');
    if (p) {
        p++;
    }
    else {
        p = (void *)filename;
    }
    if (sscanf(p, "%08X%08X%08X", &tli, &xlogid, &segment_id) != 3) {
        fprintf(logfile, "WAL segment file \"%s\" is not valid.\n", filename);
    }
    else {
        fprintf(logfile, 
                "WAL segment file %s: tli:%d, xlogid:%d, segment_id:%d.\n",
                filename, tli, xlogid, segment_id);
    }
    return;
}

void log_page_header(int num, XLogPageHeader page)
{
    CK_LOGGING;
    fprintf(logfile, 
            "=[%04d]==================================================\n", 
            num);
    fprintf(logfile, "PAGE: xlp_magic=%02X\n", page->xlp_magic);
    fprintf(logfile, "PAGE: xlp_info=%02X\n", page->xlp_info);
    fprintf(logfile, "PAGE: xlp_tli=%u\n", page->xlp_tli);
    fprintf(logfile, "PAGE: xlogid=%u\n", page->xlp_pageaddr.xlogid);
    fprintf(logfile, "PAGE: xrecoff=%u\n", page->xlp_pageaddr.xrecoff);
    if (page->xlp_info & XLP_FIRST_IS_CONTRECORD)
    {
        XLogContRecord *cont =
            (XLogContRecord *)((char *)page + XLogPageHeaderSize(page));
        fprintf(logfile, "PAGE: rem_len=%u\n", cont->xl_rem_len);
    }
    fprintf(logfile, 
            "=========================================================\n");
}

void logXLogRecord(XLogRecPtr *ptr, size_t off, XLogRecord *record)
{
    static XLogRecPtr prevRecPtr = { 0, 0};
    char *rm_name;
    char *rec_name;

    CK_LOGGING;

    fprintf(logfile, "%u/%08X: prv %u/%08X",
           ptr->xlogid, (uint32)off + segment_id * XLOG_SEG_SIZE,
           record->xl_prev.xlogid, record->xl_prev.xrecoff);

    if (!XLByteEQ(record->xl_prev, prevRecPtr))
        fprintf(logfile, "(?)");
    prevRecPtr.xlogid = ptr->xlogid;
    prevRecPtr.xrecoff = (uint32)off + segment_id * XLOG_SEG_SIZE;

    fprintf(logfile, "; xid %u; ", record->xl_xid);
    get_record_name(record->xl_rmid, record->xl_info, &rm_name, &rec_name);
    fprintf(logfile, "RM %s(%d)", rm_name, record->xl_rmid);
    fprintf(logfile, " info %02X len %u tot_len %u :%s, ", record->xl_info,
            record->xl_len, record->xl_tot_len, rec_name);
    fflush(logfile);
}

void logCompression(int compressed)
{
    CK_LOGGING;
    fprintf(logfile, "compressed: %d.\n", compressed);
    fflush(logfile);
}

void logMessage(char *msg)
{
    CK_LOGGING;
    fprintf(logfile, "%s", msg);
    fflush(logfile);
}

void log_BlkBoundary()
{
    CK_LOGGING;
    fprintf(logfile, "Found BLK boundary.\n");
    fflush(logfile);
}

static void get_record_name(int xl_rmid, int info, char **rmid_name, char **rec_name)
{
	switch(xl_rmid) {
    case RM_XLOG_ID:
        /* catalog/pg_control.h */
        *rmid_name = "RM_XLOG_ID";
        switch(info) {
        case XLOG_CHECKPOINT_SHUTDOWN:
            *rec_name = "XLOG_CHECKPOINT_SHUTDOWN";
            break;
        case XLOG_CHECKPOINT_ONLINE:
            *rec_name = "XLOG_CHECKPOINT_ONLINE";
            break;
        case XLOG_NOOP:
            *rec_name = "XLOG_NOOP";
            break;
        case XLOG_NEXTOID:
            *rec_name = "XLOG_NEXTOID";
            break;
        case XLOG_SWITCH:
            *rec_name = "XLOG_SWITCH";
            break;
        /* New to PostgreSQL 9.0 */
        case XLOG_BACKUP_END:
            *rec_name = "XLOG_BACKUP_END";
            break;
        /* New to PostgreSQL 9.0 */
        case XLOG_PARAMETER_CHANGE:
            *rec_name = "XLOG_PARAMETER_CHANGE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_XACT_ID:
        /* access/xact.h */
        *rmid_name = "RM_XACT_ID";
        switch(info) {
        case XLOG_XACT_COMMIT:
            *rec_name = "XLOG_XACT_COMMIT";
            break;
        case XLOG_XACT_PREPARE:
            *rec_name = "XLOG_XACT_PREPARE";
            break;
        case XLOG_XACT_ABORT:
            *rec_name = "XLOG_XACT_ABORT";
            break;
        case XLOG_XACT_COMMIT_PREPARED:
            *rec_name = "XLOG_XACT_COMMIT_PREPARED";
            break;
        case XLOG_XACT_ABORT_PREPARED:
            *rec_name = "XLOG_XACT_ABORT_PREPARED";
            break;
        /* New to PostgreSQL 9.0 */
        case XLOG_XACT_ASSIGNMENT:
            *rec_name = "XLOG_XACT_ASSIGNMENT";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_SMGR_ID:
        /* 
         * Definition of info value is found in src/backend/catalog/storage.c,
         * not in conventional .h file.   Maybe this is just for internal
         * use.   Here, the definition is taken from .c file.
         */
#define XLOG_SMGR_CREATE    0x10
#define XLOG_SMGR_TRUNCATE  0x20
        *rmid_name = "RM_SMGR_ID";
        switch(info) {
        case XLOG_SMGR_CREATE:
            *rec_name = "XLOG_SMGR_CREATE";
            break;
        case XLOG_SMGR_TRUNCATE:
            *rec_name = "XLOG_SMGR_TRUNCATE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_CLOG_ID:
        /* access/clog.h */
        *rmid_name = "RM_CLOG_ID";
        switch(info) {
        case CLOG_ZEROPAGE:
            *rec_name = "CLOG_ZEROPAGE";
            break;
        case CLOG_TRUNCATE:
            *rec_name = "CLOG_TRUNCATE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_DBASE_ID:
        /* commands/dbcommands.h */
        *rmid_name = "RM_DBASE_ID";
        switch (info) {
        case XLOG_DBASE_CREATE:
            *rec_name = "XLOG_DBASE_CREATE";
            break;
        case XLOG_DBASE_DROP:
            *rec_name = "XLOG_DBASE_DROP";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_TBLSPC_ID:
        /* commands/tablespace.h */
        *rmid_name = "RM_TBLSPC_ID";
        switch(info){
        case XLOG_TBLSPC_CREATE:
            *rec_name = "XLOG_TBLSPC_CREATE";
            break;
        case XLOG_TBLSPC_DROP:
            *rec_name = "XLOG_TBLSPC_DROP";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_MULTIXACT_ID:
        /* access/multixact.h */
        *rmid_name = "RM_MULTIXACT_ID";
        switch(info){
        case XLOG_MULTIXACT_ZERO_OFF_PAGE:
            *rec_name = "XLOG_MULTIXACT_ZERO_OFF_PAGE";
            break;
        case XLOG_MULTIXACT_ZERO_MEM_PAGE:
            *rec_name = "XLOG_MULTIXACT_ZERO_MEM_PAGE";
            break;
        case XLOG_MULTIXACT_CREATE_ID:
            *rec_name = "XLOG_MULTIXACT_CREATE_ID";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    /* New to PostgreSQL 9.0 */
    case RM_RELMAP_ID:
        /* utils/relmapper.h */
        *rmid_name = "RM_RELMAP_ID";
        switch(info){
        case XLOG_RELMAP_UPDATE:
            *rec_name = "XLOG_RELMAP_UPDATE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
    /* New to PostgreSQL 9.0 */
    case RM_STANDBY_ID:
        /* storage/standby.h */
        *rmid_name = "RM_STANDBY_ID";
        switch(info) {
        case XLOG_STANDBY_LOCK:
            *rec_name = "XLOG_STANDBY_LOCK";
            break;
        case XLOG_RUNNING_XACTS:
            *rec_name = "XLOG_STANDBY_XACTS";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
    case RM_HEAP2_ID:
        /* access/htup.h */
        *rmid_name = "RM_HEAP2_ID";
        info &= XLOG_HEAP_OPMASK;
        switch(info){
        case XLOG_HEAP2_FREEZE:
            *rec_name = "XLOG_HEAP2_FREEZE";
            break;
        case XLOG_HEAP2_CLEAN:
            *rec_name = "XLOG_HEAP2_CLEAN";
            break;
        /* Removed in PostgreSQL 9.0
        case XLOG_HEAP2_CLEAN_MOVE:
            *rec_name = "XLOG_HEAP2_CLEAN_MOVE";
            break;
        */
        /* New to PostgreSQL 9.0 */
        case XLOG_HEAP2_CLEANUP_INFO:
            *rec_name = "XLOG_HEAP2_CLEANUP_INFO";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
	case RM_HEAP_ID:
        /* access/htup.h */
        *rmid_name = "RM_HEAP_ID";
        info &= XLOG_HEAP_OPMASK;
        switch(info) {
        case XLOG_HEAP_INSERT:
            *rec_name = "XLOG_HEAP_INSERT";
            break;
        case XLOG_HEAP_DELETE:
            *rec_name = "XLOG_HEAP_DELETE";
            break;
        case XLOG_HEAP_UPDATE:
            *rec_name = "XLOG_HEAP_UPDATE";
            break;
        /* XLOG_HEAP_MOVE was removed in PostgreSQL 9.0.
        case XLOG_HEAP_MOVE:
            *rec_name = "XLOG_HEAP_MOVE";
            break;
        */
        case XLOG_HEAP_HOT_UPDATE:
            *rec_name = "XLOG_HEAP_HOT_UPDATE";
            break;
        case XLOG_HEAP_NEWPAGE:
            *rec_name = "XLOG_HEAP_NEWPAGE";
            break;
        case XLOG_HEAP_LOCK:
            *rec_name = "XLOG_HEAP_LOCK";
            break;
        case XLOG_HEAP_INPLACE:
            *rec_name = "XLOG_HEAP_INPLACE";
            break;
        case XLOG_HEAP_INIT_PAGE:
            *rec_name = "XLOG_HEAP_INIT_PAGE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_BTREE_ID:
        /* access/nbtree.h */
        *rmid_name = "RM_BTREE_ID";
        info &= INFO_MASK;
        switch(info){
        case XLOG_BTREE_INSERT_LEAF:
            *rec_name = "XLOG_BTREE_INSERT_LEAF";
            break;
        case XLOG_BTREE_INSERT_UPPER:
            *rec_name = "XLOG_BTREE_INSERT_UPPER";
            break;
        case XLOG_BTREE_INSERT_META:
            *rec_name = "XLOG_BTREE_INSERT_META";
            break;
        case XLOG_BTREE_SPLIT_L:
            *rec_name = "XLOG_BTREE_SPLIT_L";
            break;
        case XLOG_BTREE_SPLIT_R:
            *rec_name = "XLOG_BTREE_SPLIT_R";
            break;
        case XLOG_BTREE_SPLIT_L_ROOT:
            *rec_name = "XLOG_BTREE_SPLIT_L_ROOT";
            break;
        case XLOG_BTREE_SPLIT_R_ROOT:
            *rec_name = "XLOG_BTREE_SPLIT_R_ROOT";
            break;
        case XLOG_BTREE_DELETE:
            *rec_name = "XLOG_BTREE_DELETE";
            break;
        case XLOG_BTREE_DELETE_PAGE:
            *rec_name = "XLOG_BTREE_DELETE_PAGE";
            break;
        case XLOG_BTREE_DELETE_PAGE_META:
            *rec_name = "XLOG_BTREE_DELETE_PAGE_META";
            break;
        case XLOG_BTREE_NEWROOT:
            *rec_name = "XLOG_BTREE_NEWROOT";
            break;
        case XLOG_BTREE_DELETE_PAGE_HALF:
            *rec_name = "XLOG_BTREE_DELETE_PAGE_HALF";
            break;
        /* New to PostgreSLQ 9.0 */
        case XLOG_BTREE_VACUUM:
            *rec_name = "XLOG_BTREE_VACUUM";
            break;
        /* New to PostgreSLQ 9.0 */
        case XLOG_BTREE_REUSE_PAGE:
            *rec_name = "XLOG_BTREE_REUSE_PAGE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_HASH_ID:
        /* Now hash does not write any WAL */
        *rmid_name = "RM_HASH_ID";
        *rec_name = "Unknown";
        return;
    case RM_GIN_ID:
        /* access/gin.h */
        *rmid_name = "RM_GIN_ID";
        info &= INFO_MASK;
        switch(info){
        case XLOG_GIN_CREATE_INDEX:
            *rec_name = "XLOG_GIN_CREATE_INDEX";
            break;
        case XLOG_GIN_CREATE_PTREE:
            *rec_name = "XLOG_GIN_CREATE_PTREE";
            break;
        case XLOG_GIN_INSERT:
            *rec_name = "XLOG_GIN_INSERT";
            break;
        case XLOG_GIN_SPLIT:
            *rec_name = "XLOG_GIN_SPLIT";
            break;
        case XLOG_GIN_VACUUM_PAGE:
            *rec_name = "XLOG_GIN_VACUUM_PAGE";
            break;
        case XLOG_GIN_DELETE_PAGE:
            *rec_name = "XLOG_GIN_DELETE_PAGE";
            break;
        case XLOG_GIN_UPDATE_META_PAGE:
            *rec_name = "XLOG_GIN_UPDATE_META_PAGE";
            break;
        case XLOG_GIN_INSERT_LISTPAGE:
            *rec_name = "XLOG_GIN_INSERT_LISTPAGE";
            break;
        case XLOG_GIN_DELETE_LISTPAGE:
            *rec_name = "XLOG_GIN_DELETE_LISTPAGE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_GIST_ID:
        /* access/gist_private.h */
        *rmid_name = "RM_GIST_ID";
        info &= INFO_MASK;
        switch(info){
        case XLOG_GIST_PAGE_UPDATE:
            *rec_name = "XLOG_GIST_PAGE_UPDATE";
            break;
        case XLOG_GIST_NEW_ROOT:
            *rec_name = "XLOG_GIST_NEW_ROOT";
            break;
        case XLOG_GIST_PAGE_SPLIT:
            *rec_name = "XLOG_GIST_PAGE_SPLIT";
            break;
        case XLOG_GIST_INSERT_COMPLETE:
            *rec_name = "XLOG_GIST_INSERT_COMPLETE";
            break;
        case XLOG_GIST_CREATE_INDEX:
            *rec_name = "XLOG_GIST_CREATE_INDEX";
            break;
        case XLOG_GIST_PAGE_DELETE:
            *rec_name = "XLOG_GIST_PAGE_DELETE";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    case RM_SEQ_ID:
        /* commands/sequence.h */
        *rmid_name = "RM_SEQ_ID";
        switch(info){
        case XLOG_SEQ_LOG:
            *rec_name = "XLOG_SEQ_LOG";
            break;
        default:
            *rec_name = "Unknown";
            break;
        }
        return;
    default:
        /* Error ... perhaps */
        *rmid_name = "Unknown";
        *rec_name = "Unknown";
        return;
    }
}
